// Copyright Istio Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package mcp

import (
	"errors"
	"fmt"
	"strings"
	"sync"
	"time"

	"github.com/gogo/protobuf/types"

	"istio.io/api/label"
	"istio.io/istio/pilot/pkg/model"
	"istio.io/istio/pkg/config"
	"istio.io/istio/pkg/config/schema/collection"
	"istio.io/istio/pkg/config/schema/collections"
	"istio.io/istio/pkg/config/schema/gvk"
	"istio.io/istio/pkg/mcp/sink"
	"istio.io/pkg/log"
)

var (
	errUnsupported = errors.New("this operation is not supported by mcp controller")
)

// Controller is a combined interface for ConfigStoreCache
// and MCP Updater
type Controller interface {
	model.ConfigStoreCache
	sink.Updater
}

// Options stores the configurable attributes of a Control
type Options struct {
	DomainSuffix string
	XDSUpdater   model.XDSUpdater
	Revision     string
}

// controller is a temporary storage for the changes received
// via MCP server
type controller struct {
	configStoreMu sync.RWMutex
	// keys [type][namespace][name]
	configStore   map[config.GroupVersionKind]map[string]map[string]*config.Config
	options       *Options
	eventHandlers map[config.GroupVersionKind][]func(config.Config, config.Config, model.Event)

	syncedMu sync.Mutex
	synced   map[string]bool
}

// NewController provides a new Controller controller
func NewController(options *Options) Controller {
	synced := make(map[string]bool)

	for _, n := range collections.Pilot.CollectionNames() {
		synced[n.String()] = false
	}

	return &controller{
		configStore:   make(map[config.GroupVersionKind]map[string]map[string]*config.Config),
		options:       options,
		eventHandlers: make(map[config.GroupVersionKind][]func(config.Config, config.Config, model.Event)),
		synced:        synced,
	}
}

func (c *controller) Schemas() collection.Schemas {
	return collections.Pilot
}

// List returns all the config that is stored by type and namespace
// if namespace is empty string it returns config for all the namespaces
func (c *controller) List(typ config.GroupVersionKind, namespace string) (out []config.Config, err error) {
	_, ok := collections.Pilot.FindByGroupVersionKind(typ)
	if !ok {
		return nil, fmt.Errorf("list unknown type %s", typ)
	}
	c.configStoreMu.Lock()
	byType, ok := c.configStore[typ]
	c.configStoreMu.Unlock()
	if !ok {
		return nil, nil
	}

	if namespace == "" {
		// ByType does not need locking since
		// we replace the entire sub-map
		for _, byNamespace := range byType {
			for _, config := range byNamespace {
				out = append(out, *config)
			}
		}
		return out, nil
	}

	for _, config := range byType[namespace] {
		out = append(out, *config)
	}
	return out, nil
}

// Apply receives changes from MCP server and creates the
// corresponding config
func (c *controller) Apply(change *sink.Change) error {
	s, ok := collections.Pilot.Find(change.Collection)
	if !ok {
		return fmt.Errorf("apply type not supported %s", change.Collection)
	}

	kind := s.Resource().GroupVersionKind()
	configUpdated := map[model.ConfigKey]struct{}{} // If non-incremental, we use empty configUpdated to indicate all.

	// innerStore is [namespace][name]
	innerStore := make(map[string]map[string]*config.Config)
	for _, obj := range change.Objects {
		namespace, name := extractNameNamespace(obj.Metadata.Name)

		createTime := time.Now()
		if obj.Metadata.CreateTime != nil {
			var err error
			if createTime, err = types.TimestampFromProto(obj.Metadata.CreateTime); err != nil {
				// Do not return an error, instead discard the resources so that Pilot can process the rest.
				log.Warnf("Discarding incoming MCP resource: invalid resource timestamp (%s/%s): %v", namespace, name, err)
				continue
			}
		}

		if change.Incremental {
			configUpdated[model.ConfigKey{
				Kind:      kind,
				Name:      name,
				Namespace: namespace,
			}] = struct{}{}
		}
		conf := &config.Config{
			Meta: config.Meta{
				GroupVersionKind:  kind,
				Name:              name,
				Namespace:         namespace,
				ResourceVersion:   obj.Metadata.Version,
				CreationTimestamp: createTime,
				Labels:            obj.Metadata.Labels,
				Annotations:       obj.Metadata.Annotations,
				Domain:            c.options.DomainSuffix,
			},
			Spec: obj.Body,
		}

		if _, err := s.Resource().ValidateConfig(*conf); err != nil {
			// Do not return an error, instead discard the resources so that Pilot can process the rest.
			log.Warnf("Discarding incoming MCP resource: validation failed (%s/%s): %v", conf.Namespace, conf.Name, err)
			continue
		}

		// Skip configs not selected by this revision
		if c.objectInRevision(conf) {
			namedConfig, ok := innerStore[conf.Namespace]
			if ok {
				namedConfig[conf.Name] = conf
			} else {
				innerStore[conf.Namespace] = map[string]*config.Config{
					conf.Name: conf,
				}
			}
		}
	}
	for _, removed := range change.Removed {
		namespace, name := extractNameNamespace(removed)
		configUpdated[model.ConfigKey{
			Kind:      kind,
			Name:      name,
			Namespace: namespace,
		}] = struct{}{}
	}

	var prevStore map[string]map[string]*config.Config

	c.configStoreMu.Lock()
	prevStore = c.configStore[kind]
	// Incremental update when received incremental change
	if change.Incremental {

		//Although it is not a deep copy, there is no problem because the config will not be modified
		prevCache := make(map[string]map[string]*config.Config, len(prevStore))
		for namespace, namedConfig := range prevStore {
			prevCache[namespace] = make(map[string]*config.Config, len(namedConfig))
			for name, config := range namedConfig {
				prevCache[namespace][name] = config
			}
		}
		prevStore = prevCache

		c.removeConfig(kind, change.Removed)
		c.incrementalUpdate(kind, innerStore)
		innerStore = c.configStore[kind]
	}
	c.configStore[kind] = innerStore
	c.configStoreMu.Unlock()
	c.sync(change.Collection)

	if kind == gvk.ServiceEntry {
		c.serviceEntryEvents(innerStore, prevStore)
	} else if c.options.XDSUpdater != nil {
		c.options.XDSUpdater.ConfigUpdate(&model.PushRequest{
			Full:           true,
			ConfigsUpdated: configUpdated,
			Reason:         []model.TriggerReason{model.ConfigUpdate},
		})
	}

	return nil
}

func (c *controller) objectInRevision(o *config.Config) bool {
	configEnv, f := o.Labels[label.IstioRev]
	if !f {
		// This is a global object, and always included
		return true
	}
	// Otherwise, only return if the
	return configEnv == c.options.Revision
}

// HasSynced returns true if the first batch of items has been popped
func (c *controller) HasSynced() bool {
	var notReady []string

	c.syncedMu.Lock()
	for messageName, synced := range c.synced {
		if !synced {
			notReady = append(notReady, messageName)
		}
	}
	c.syncedMu.Unlock()

	if len(notReady) > 0 {
		return false
	}

	log.Infof("Configuration synced")
	return true
}

// RegisterEventHandler registers a handler using the type as a key
func (c *controller) RegisterEventHandler(kind config.GroupVersionKind, handler func(config.Config, config.Config, model.Event)) {
	c.eventHandlers[kind] = append(c.eventHandlers[kind], handler)
}

// Run is not implemented
func (c *controller) Run(<-chan struct{}) {
	log.Warnf("Run: %s", errUnsupported)
}

// Get is not implemented
func (c *controller) Get(_ config.GroupVersionKind, _, _ string) *config.Config {
	log.Warnf("get %s", errUnsupported)
	return nil
}

// Update is not implemented
func (c *controller) Update(config.Config) (newRevision string, err error) {
	log.Warnf("update %s", errUnsupported)
	return "", errUnsupported
}

func (c *controller) UpdateStatus(config.Config) (newRevision string, err error) {
	log.Warnf("updateStatus %s", errUnsupported)
	return "", errUnsupported
}

func (c *controller) Patch(orig config.Config, patchFn config.PatchFunc) (string, error) {
	log.Warnf("patch %s", errUnsupported)
	return "", errUnsupported
}

// Create is not implemented
func (c *controller) Create(config.Config) (revision string, err error) {
	log.Warnf("create %s", errUnsupported)
	return "", errUnsupported
}

// Delete is not implemented
func (c *controller) Delete(_ config.GroupVersionKind, _, _ string, _ *string) error {
	return errUnsupported
}

func (c *controller) sync(collection string) {
	c.syncedMu.Lock()
	c.synced[collection] = true
	c.syncedMu.Unlock()
}

func (c *controller) serviceEntryEvents(currentStore, prevStore map[string]map[string]*config.Config) {
	dispatch := func(prev config.Config, model config.Config, event model.Event) {}
	if handlers, ok := c.eventHandlers[gvk.ServiceEntry]; ok {
		dispatch = func(prev config.Config, config config.Config, event model.Event) {
			log.Debugf("MCP event dispatch: key=%v event=%v", config.Key(), event.String())
			for _, handler := range handlers {
				handler(prev, config, event)
			}
		}
	}

	// add/update
	for namespace, byName := range currentStore {
		for name, cfg := range byName {
			if prevByNamespace, ok := prevStore[namespace]; ok {
				if prevConfig, ok := prevByNamespace[name]; ok {
					if cfg.ResourceVersion != prevConfig.ResourceVersion {
						dispatch(*prevConfig, *cfg, model.EventUpdate)
					}
				} else {
					dispatch(config.Config{}, *cfg, model.EventAdd)
				}
			} else {
				dispatch(config.Config{}, *cfg, model.EventAdd)
			}
		}
	}

	// delete
	for namespace, prevByName := range prevStore {
		for name, prevConfig := range prevByName {
			if _, ok := currentStore[namespace][name]; !ok {
				dispatch(config.Config{}, *prevConfig, model.EventDelete)
			}
		}
	}
}

func extractNameNamespace(metadataName string) (string, string) {
	segments := strings.Split(metadataName, "/")
	if len(segments) == 2 {
		return segments[0], segments[1]
	}
	return "", segments[0]
}

func (c *controller) removeConfig(kind config.GroupVersionKind, resources []string) {
	for _, fullName := range resources {
		namespace, name := extractNameNamespace(fullName)
		if byType, ok := c.configStore[kind]; ok {
			if byNamespace, ok := byType[namespace]; ok {
				if conf, ok := byNamespace[name]; ok {
					delete(byNamespace, conf.Name)
				}
				// clear parent map also
				if len(byNamespace) == 0 {
					delete(byType, namespace)
				}
			}
			// clear parent map also
			if len(byType) == 0 {
				delete(c.configStore, kind)
			}
		}
	}
}

func (c *controller) incrementalUpdate(kind config.GroupVersionKind, conf map[string]map[string]*config.Config) {
	if len(conf) == 0 {
		return
	}
	if byType, ok := c.configStore[kind]; ok {
		for namespace, namedConf := range conf {
			if byNamespace, ok := byType[namespace]; ok {
				for name, config := range namedConf {
					byNamespace[name] = config
				}
			} else {
				byType[namespace] = namedConf
			}
		}
	} else {
		c.configStore[kind] = conf
	}
}
